/*
  d. bodnar  revised 6-16-2016 with Rotary Encoder
  modified msport 13-03-2020 for 1.44" 128x128 SPI TFT display or 1.8" 128x160
  modified msport for 4x4 keyboard to drive points and accessories
  library ADAFRUIT
  Version  ST7735 from graphictest example
  encoder library with reset
  https://forum.arduino.cc/index.php?topic=528379.0 eeprom
  https://byfeel.info/eeprom-ou-spiffs/
  02-01-2021 (C) Locoduino licence Open Software
  modified msport 17-06-2022 for ESP32 pcb ESP_PAD_22 OK
  display functions revisited 23 08 2023
*/
#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library
#include <TFT_eSPI.h> // Graphics and font library for ST7735 driver chip
#include <SPI.h>
#include <ESP32Encoder.h>
ESP32Encoder encoder; // revoir tout encoder
int encoderCount = 0 ;
int oldEncoderCount = 1;

TFT_eSPI tft = TFT_eSPI();  // Invoke library, pins defined in User_Setup.h

/* 
  Make sure all the display driver and pin comnenctions are correct by
  editting the User_Setup.h file in the TFT_eSPI library folder.
 #########################################################################
 ###### DON'T FORGET TO UPDATE THE User_Setup.h FILE IN THE LIBRARY ######
 #########################################################################
 *  voir  User_Setup.h file in the TFT_eSPI lib esp32
#define ST7735_DRIVER      // Define additional parameters below for this display
 #define TFT_HEIGHT 128
 #define TFT_WIDTH  128
 #define ST7735_GREENTAB3
 #define TFT_MOSI 13 
 #define TFT_SCLK 14 
 #define TFT_CS   15  
 #define TFT_DC   12  
 #define TFT_RST   -1  //  to the Arduino reset, set this #define pin to -1!

Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS,  TFT_DC, TFT_RST);
//Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);
 */
#include<EEPROM.h>

// Color definitions
#define BLACK   0x0000
#define BLUE    0x001F
#define RED     0xF800
#define GREEN   0x07E0
#define CYAN    0x07FF
#define MAGENTA 0xF81F
#define YELLOW  0xFFE0  
#define WHITE   0xFFFF


const int batPin =  33;      // battery sur pcb pad_22
const int tempPin =  5;      // CTN sur pcb pad_22
int buttonPin = 26; // button on rotary
int positionSpeed; // keep encoder position for speed
int positionPoint; // keep encoder position for Point numberTurntable
int positionTurntable; // keep encoder position for Turntable
int positionAccessory; // keep encoder position for Accessory number
int buttonState = 0;
int encoderChange = 0; // flag to show encoder changed

byte Key;
char key ;
byte Fx = 0;
// Array set for 4 Loco2 - change the 7 numbers in the next 7 declarations
int maxLocos = 4; // number of loco addresses
int LocoAddress[4] = {3, 36, 35, 12};
int LocoDirection[4] = {1, 1, 1, 1};
int LocoSpeed[4] = {0, 0, 0, 0};
byte LocoFN[4][4]= {{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0},{0, 0, 0, 0}} ; // [LIGNES][COLONNES] for 13-20 and 21-28 2&3 groups / loco addresses
byte LocoFN0to4[4]= {0, 0, 0, 0}; // 0 group
byte LocoFN5to8[4]= {0, 0, 0, 0}; // 0 group
byte LocoFN9to12[4]= {0, 0, 0, 0}; // 1 group
byte LocoFN13to20[4]= {0, 0, 0, 0}; // 2 group
byte LocoFN21to28[4]= {0, 0, 0, 0}; // 3 group
byte group[4]= {0, 0, 0, 0}; // 9-24 in progress
int delta[4] = {0,8,12,20};  

boolean AccessoryDone = false;
int PointAddressDble = 2;
int AccessoryAddressDble = 2;
int TurntableAddressDble = 2;
int xxxxx = 0;
int counter = 0;

float vBat = 3.7;
int ZeroSpeedFlag = 0;
int ActiveAddress = 0; // make address1 active

int z = 0;
int powerTemp = 0;
int i = 0;
#include <Keypad.h>
const byte ROWS = 4; //four rows
const byte COLS = 4; //four columns
char keys[ROWS][COLS] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};
byte rowPins[ROWS] = {19, 18, 17, 25}; // 25, 17, 18, 19 //connect to the row pinouts of the keypad
byte colPins[COLS] = {21, 22, 23, 27}; //  //connect to the column pinouts of the keypad
Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );
unsigned long previousMillis = 0;        // during setup
unsigned long delayed = 4000 ;           // during setup

int debug = 0; // set to 1 to show debug info on serial port - may cause issues with DCC++ depending on what is sent

struct LocoStructure {
  unsigned long magic;
  byte struct_version;
  uint16_t checksum;
  int address1; // adresse de la loco 1
  int address2; // adresse de la loco 2
  int address3; // adresse de la loco 3
  int address4; // adresse de la loco 4
};
struct LocoStructure ms {123456789, 4, 0, 8, 36,  12, 68 } ;

void setup() {   // =========================== SETUP ========================
  EEPROM.begin(128);
  pinMode(buttonPin, INPUT_PULLUP);
  Serial.begin(115200);
  delay(200);
  ESP32Encoder::useInternalWeakPullResistors=UP;   // Enable the weak pull up resistors
  encoder.attachSingleEdge(16, 4);  // Attache pins for use as encoder pins
  encoder.setCount(0);    // set starting count value after attaching  
  tft.init(); 
  tft.setRotation ( 1 );
  tft.setTextWrap(false);
  tft.fillScreen(TFT_BLACK);
  tft.setCursor(0, 0);  // col, ligne 0, 30
  tft.setTextColor(TFT_GREEN, TFT_BLACK);
  tft.setTextSize(2);  // 6 lines 10 char size 2 portrait
  getAddresses();  // read eeprom
  tft.println(F("Appuyer:"));
  tft.println(F("changer "));
  tft.println(F("Nb locos")); 
  tft.setTextColor(TFT_YELLOW);
  tft.println(maxLocos);
  pinMode(batPin, INPUT);              // to get battery voltage on A7
  vBat = analogRead(batPin);
  vBat = vBat / 607 ;
  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  tft.println(F("Batterie :")); 
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.print(vBat, 1); 
  tft.println(F(" V")); 
  delay(2000);
  buttonState = digitalRead(buttonPin);
  if (debug == 1) Serial.println(buttonState);
  if (buttonState == LOW) {
  if (debug == 1) Serial.println("getNumberOfLocos");
    getNumberOfLocos();
  }
  tft.setCursor(0, 0);  // col, ligne 0, 0
  tft.fillScreen(TFT_BLACK);
  tft.setTextColor(TFT_GREEN);
  tft.println(F("DCC++"));
  tft.println(F("Throttle"));
  tft.println(F("E_ST_13"));  
  tft.println(F("23-08-2023")); 
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  do {  ;
  }      while (millis() < previousMillis + delayed );   
  if (debug == 1) Serial.println("E_ST_1 setup");
  Serial.print(F("<1>"));  // power ON to DCC++ unit
  delay(1500);
  tft.fillScreen(TFT_BLACK);
  doMainLCD(); //
}  // END SETUP

void getNumberOfLocos() {  // initial choice of loco number to cycle in
  //maxLocos = 4;// number of loco addresses
  tft.fillScreen(TFT_BLACK);  // clears the screen and buffer 
  tft.setCursor(0, 0);
  tft.setTextColor(TFT_WHITE);
  tft.println(F("Nb locos")); tft.print(F("  = ")); 
  tft.setTextColor(TFT_CYAN); tft.print(maxLocos);
  tft.println("");
  tft.setTextColor(TFT_WHITE);
  tft.println(F("Nouveau :")); 
  tft.print(F(" 1-4 : "));  
  tft.setTextColor(TFT_CYAN);
  do {
    key = keypad.getKey();
    if (debug == 1) Serial.print("key = ");
    if (debug == 1) Serial.println(key);
    key = key - 48 - 1; // convert from ascii value
    if (key == 1 || key == 2 || key == 3 || key == 4) { // 1-4
      maxLocos = key + 1;
      if (debug == 1)    Serial.print(F("new maxLocos = "));
      if (debug == 1)  Serial.println(maxLocos);
      tft.print(maxLocos);
      delay(1400);
      saveAddresses();
      break;
    }
  } while (key != 1 || key != 2 || key != 3 || key != 4);
}  // END  getNumberOfLocos

void loop() { // =========================== LOOP ========================
  key = keypad.getKey();
  if (key == 42) { // *
    all2ZeroSpeeed();
    getLocoAddress();
    key = 0;
  }
  // ROUTINE USING ROTARY ENCODER FOR SPEED AND BUTTON ON IT FOR DIRECTION
  getEncoder();  
  if (encoderChange == 1) {
    encoderChange = 0;
    LocoSpeed[ActiveAddress] = encoderCount;
    doDCC();
    doMainLCD();    
  }
  buttonState = digitalRead(buttonPin);
  if (buttonState == LOW) {
    delay(50);
    buttonState = digitalRead(buttonPin); // check a 2nd time to be sure
    if (buttonState == LOW) {// check a 2nd time to be sure
      LocoDirection[ActiveAddress] = !LocoDirection[ActiveAddress];
      doDCC();
      doMainLCD();
      do {  // routine to stay here till button released & not toggle direction
        buttonState = digitalRead(buttonPin);
      }      while (buttonState == LOW);
    }
  }
  if (key == 35) { // #
    ActiveAddress++;
    if (ActiveAddress >= maxLocos) ActiveAddress = 0;
    tft.setCursor (60, 40);       // col, line
    tft.print("     ");    
    doMainLCD();
    delay(200);
    key = 0;
    encoder.setCount(LocoSpeed[ActiveAddress]);
    doDCC();
  }
  if (key != 42 && key != 35 && key >= 48 && key <= 57) { // figure or letter, not #, not * 
    doFunction();
  }
  if (key == 65) doPoint();                  // letter A only
  if (key == 66) doTurntable();               // letter B only
  if (key == 67) doAccessory();              // letter C only
  if (key == 68) doChangeGroupFunction() ;   // letter D only

}  // END LOOP

void getEncoder() {
  encoderCount = (int32_t)encoder.getCount();
  if (oldEncoderCount != encoderCount) {
    if ( encoderCount > 127 ) encoder.setCount(127) ;
    if ( encoderCount < 0 ) encoder.setCount(0) ;
    oldEncoderCount = encoderCount;
    encoderChange = 1; // flag to show a change took place
  }
}  // END GET ENCODER ROUTINE

//START DO FUNCTION BUTTONS
void doFunction() {
  key = key - 48 ; // - 1; // convert from ascii value, 48 = zero, press 1 gives F0
  tft.setCursor (84, 60);       // go to 6th line, 12 per space ex 84, one space before F
  tft.setTextColor(TFT_CYAN, TFT_BLACK);    
  tft.print("F");
  tft.setTextColor(TFT_YELLOW, TFT_BLACK);  
  tft.print(key + (delta[group[ActiveAddress]]) , DEC ); // display Function #
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  tft.print("  ");
  if (debug == 1) Serial.print("got a keypad button ");
  if (debug == 1) Serial.println(key + (8 * group[ActiveAddress]), DEC);
    
switch (group[ActiveAddress]) {
  case 0:
  if (key == 0) {
    if (bitRead(LocoFN0to4[ActiveAddress], 4) == 0 ) { ///
      bitWrite(LocoFN0to4[ActiveAddress], 4, 1);
    }
    else {
      if (bitRead(LocoFN0to4[ActiveAddress], 4) == 1 ) {
        bitWrite(LocoFN0to4[ActiveAddress], 4, 0);
      }
    }      
    doDCCfunction04();
  }
  if (key > 0 && key <= 4) {
    if (bitRead(LocoFN0to4[ActiveAddress], key - 1  ) == 0 ) { ///
      bitWrite(LocoFN0to4[ActiveAddress], key - 1 , 1);
    }
    else {
      if (bitRead(LocoFN0to4[ActiveAddress], key - 1 ) == 1 ) {
        bitWrite(LocoFN0to4[ActiveAddress], key - 1 , 0);
      }
    }
    doDCCfunction04();
    if (debug == 1) Serial.print(LocoFN0to4[ActiveAddress], BIN);
    if (debug == 1) Serial.println(" LocoFN0to4[ActiveAddress] d ");
    if (debug == 1) Serial.print(LocoFN0to4[ActiveAddress], DEC);
    if (debug == 1) Serial.println(" LocoFN0to4[ActiveAddress]");
  }
  if (key >= 5 && key <= 8) {
    key = key - 5;
    if (bitRead(LocoFN5to8[ActiveAddress], key) == 0 ) {
      bitWrite(LocoFN5to8[ActiveAddress], key, 1);
    }
    else {
      if (bitRead(LocoFN5to8[ActiveAddress], key) == 1 ) {
        bitWrite(LocoFN5to8[ActiveAddress], key, 0);
      }
    }
    doDCCfunction58();
    if (debug == 1) Serial.print(LocoFN5to8[ActiveAddress], BIN);
    if (debug == 1) Serial.println(" LocoFN5to8[ActiveAddress] d ");
    if (debug == 1) Serial.print(LocoFN5to8[ActiveAddress], DEC);
    if (debug == 1) Serial.println("LocoFN5to8[ActiveAddress]");
  }
  break;
  case 1: // 9-12
  if (key >= 1 && key <= 4) {
    key = key - 1;
    if (bitRead(LocoFN9to12[ActiveAddress], key) == 0 ) {
      bitWrite(LocoFN9to12[ActiveAddress], key, 1);
    }
    else {
      if (bitRead(LocoFN9to12[ActiveAddress], key) == 1 ) {
        bitWrite(LocoFN9to12[ActiveAddress], key, 0);
      }
    }
    doDCCfunction912();
  }
  break;
  case 2: // 13-20
  if (key >= 0 && key <= 8)  {
    key = key - 1;
    if (bitRead(LocoFN13to20[ActiveAddress], key) == 0 ) {
      bitWrite(LocoFN13to20[ActiveAddress], key, 1);
    }
    else {
      if (bitRead(LocoFN13to20[ActiveAddress], key) == 1 ) {
        bitWrite(LocoFN13to20[ActiveAddress], key, 0);
      }
    }
    doDCCfunction1320();
  }
  break;
  case 3: // 21-28
  if (key >= 0 && key <= 8) {
    key = key - 1;
    if (bitRead(LocoFN21to28[ActiveAddress], key) == 0 ) {
      bitWrite(LocoFN21to28[ActiveAddress], key, 1);
    }
    else {
      if (bitRead(LocoFN21to28[ActiveAddress], key) == 1 ) {
        bitWrite(LocoFN21to28[ActiveAddress], key, 0);
      }
    }
    doDCCfunction2128();
  }
  break;
  default:
    // statements
  break;
  }
  if (key == -1)
  {
  tft.setCursor (84,60);       // go to end of 6th line, 6 char ex 84,2
    tft.print("  F ");
    // key = 0;
    tft.print(key, DEC); // modif inversion des deux lignes,  = 0 comment
    LocoFN0to4[ActiveAddress] = B00000000; //clear variables for which functions are set
    LocoFN5to8[ActiveAddress] = B00000000;
    if ( group[ActiveAddress] == 0) {
      doDCCfunction04();
      doDCCfunction58();
    }
    if ( group[ActiveAddress] == 1) {  LocoFN9to12[ActiveAddress] = B00000000;  doDCCfunction912();}
    if ( group[ActiveAddress] == 2) {  LocoFN13to20[ActiveAddress] = B00000000;  doDCCfunction1320();}
    if ( group[ActiveAddress] == 3) {  LocoFN21to28[ActiveAddress] = B00000000;  doDCCfunction2128();}
    delay(500);
    key = 0;
  }
  key = 0;
  // delay(500);
  doMainLCD();
}//END DO FUNCTION BUTTONS


void doPoint() { // START DO POINT - SWITCH
 while ((key == 65 || (key == 0)) ) {  // A 65 || (key == 0)  
  key = keypad.getKey();
  doPointLCD();        
  // ROUTINE USING ROTARY ENCODER FOR POINTS AND BUTTON ON IT TO TRIGGER COMMAND
  signed char pos = (int32_t)encoder.getCount();
  if (pos != 0) {
    positionPoint += pos;
    if (positionPoint < 2) {positionPoint = 2;}
    if (positionPoint > 126) {positionPoint = 126;}
    PointAddressDble = positionPoint ; // + 2 ; // 
    encoder.setCount(0);
    doPointLCD();   
  }
  buttonState = digitalRead(buttonPin);
  if (buttonState == LOW) {
    delay(50);
    buttonState = digitalRead(buttonPin); // check a 2nd time to be sure
    if (buttonState == LOW) {// check a 2nd time to be sure
      doDCCpoint();  
      do {  // routine to stay here till button released & not triggering accessory
        buttonState = digitalRead(buttonPin); 
        tft.setCursor(0, 100); // start of 4th line
        tft.setTextColor(TFT_RED, TFT_BLACK);
        tft.print("COMMANDE   ");         
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
        tft.print(PointAddressDble / 2 , DEC);          
      }      while (buttonState == LOW);
    }
  }
 }  // end while
  if ((key != 65 && (key != 0)) ) {
    doMainLCD();
    delay(200);
    key = 0;
    doDCC(); // driving loco
    group[ActiveAddress]--; // in order to erase message
    doChangeGroupFunction();
  }
}  //  END DO POINT 

void doAccessory() { //START DO ACCESSORY
 while ((key == 67 || (key == 0)) ) {  // C 67 || (key == 0)
  key = keypad.getKey();
  doAccessoryLCD();
    // ROUTINE USING ROTARY ENCODER FOR ACCESSORIES AND BUTTON ON IT TO TRIGGER COMMAND
  signed char pos = (int32_t)encoder.getCount();
  if (pos != 0) {
    positionAccessory += pos;
    if (positionAccessory < 2) {positionAccessory = 2;}
    if (positionAccessory > 255) {positionAccessory = 255;}
    AccessoryAddressDble = positionAccessory ; // + 2 ; // 
    encoder.setCount(0);
    doAccessoryLCD();   
  }
  buttonState = digitalRead(buttonPin);
  if (buttonState == LOW) {
    delay(50);
    buttonState = digitalRead(buttonPin); // check a 2nd time to be sure
    if (buttonState == LOW) {// check a 2nd time to be sure
      doDCCaccessory();  
      do {  // routine to stay here till button released & not triggering accessory
        buttonState = digitalRead(buttonPin); 
        tft.setCursor(0, 100); // start of 5th line
        tft.setTextColor(TFT_RED, TFT_BLACK);
        tft.print("COMMANDE   ");         
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
        tft.print(AccessoryAddressDble , DEC);          
      }      while (buttonState == LOW);
    }
  }
 }  // end while
  if ((key != 67 && (key != 0)) ) {
    doMainLCD();
    delay(200);
    key = 0;
    doDCC(); // driving loco
    group[ActiveAddress]--; // in order to erase message
    doChangeGroupFunction();
  }
} // END DO ACCESSSORY


void doTurntable() { //START DO TURNTABLE
 while ((key == 66 || (key == 0)) ) {  // B 66 || (key == 0)
  key = keypad.getKey();
  doTurntableLCD();
    // ROUTINE USING ROTARY ENCODER FOR TURNTABLE AND BUTTON ON IT TO TRIGGER COMMAND
  signed char pos = (int32_t)encoder.getCount();
  if (pos != 0) {
    positionTurntable += pos;
    if (positionTurntable < 2) {positionTurntable = 2;}
    if (positionTurntable > 3) {positionAccessory = 3;}
    TurntableAddressDble = positionTurntable ; // + 2 ; // 
    encoder.setCount(0);
    doTurntableLCD();   
  }
  buttonState = digitalRead(buttonPin);
  if (buttonState == LOW) {
    delay(50);
    buttonState = digitalRead(buttonPin); // check a 2nd time to be sure
    if (buttonState == LOW) {// check a 2nd time to be sure
      doDCCturntable();  
      do {  // routine to stay here till button released & not triggering accessory
        buttonState = digitalRead(buttonPin); 
        tft.setCursor(0, 100); // start of 5th line
        tft.setTextColor(TFT_RED, TFT_BLACK);
        tft.print("COMMANDE   ");         
        tft.setTextColor(TFT_WHITE, TFT_BLACK);
      }      while (buttonState == LOW);
    }
  }
 }  // end while
  if ((key != 66 && (key != 0)) ) {
    doMainLCD();
    delay(200);
    key = 0;
    doDCC(); // driving loco
    group[ActiveAddress]--; // in order to erase message
    doChangeGroupFunction();
  }
} // END DO TURNTABLE

void showFirstLine() {
  // break;
  if (debug == 1) Serial.println(" ");
  tft.setCursor(0, 0); 
  tft.print("                ");// clear
  tft.setCursor(0, 0);
  tft.print("  L");
  tft.print(ActiveAddress + 1);
  tft.print(" = ");
  tft.print(LocoAddress[ActiveAddress]);

  for (int zzz = 0; zzz <= 3; zzz++) {
    if (debug == 1) Serial.print("add # ");
    if (debug == 1) Serial.print(zzz + 1);
    if (debug == 1) Serial.print(" ");
    if (debug == 1) Serial.println(LocoAddress[zzz]);
  }
}

void doChangeGroupFunction(){
  group[ActiveAddress]++;
  if ( group[ActiveAddress] > 3) { group[ActiveAddress]= 0; }
  tft.setCursor(60, 60);
  tft.setTextColor(TFT_CYAN, TFT_BLACK);    
  tft.print("G");
  tft.setTextColor(TFT_YELLOW, TFT_BLACK);  
  tft.print(group[ActiveAddress]); 
  if (group[ActiveAddress]== 2) { tft.setCursor(108, 80);  tft.print("2");}
  if (group[ActiveAddress]== 3) { tft.setCursor(108, 100); tft.print("3");}
  tft.setTextColor(TFT_CYAN, TFT_BLACK);  
  if (group[ActiveAddress]!= 2) { tft.setCursor(108, 80);  tft.print("2");}
  if (group[ActiveAddress]!= 3) { tft.setCursor(108, 100); tft.print("3");}
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
}

void getLocoAddress() {
  int saveAddress =   LocoAddress[ActiveAddress];
  Serial.print("<0>");// power off to DCC++ unit
  int total = 0;
  counter = 0;
  tft.fillScreen(TFT_BLACK); // 
  tft.setCursor(0, 0);
  tft.print("Adresse ");
  tft.setCursor(0, 20);
  tft.print("DCC ? ");
  tft.println(ActiveAddress + 1); 
  if (debug == 1) Serial.println(" @ top");
  do {
    key = keypad.getKey();
    if (key == '#' | key == '*') { //abort when either is hit
      //LocoAddress[ActiveAddress] = saveAddress;
      total = saveAddress;
      break;// exit routine if # button pressed - ABORT new address
    }
    if (key) {
      counter++;
      int number =  key - 48;
      total = total * 10 + number;
      if (debug == 1) Serial.print("TOTAL = ");
      if (debug == 1) Serial.println(total);
      tft.setCursor(0, 40); 
      if (total == 0) {   // print multiple zeros for leading zero number
        for (int tempx = 1; tempx <= counter; tempx++) {
          tft.print("0");
        }
      }
      else tft.print(total);
      if (debug == 1) Serial.print("Counter = ");
      if (debug == 1) Serial.print(counter);
      if (debug == 1) Serial.print("  key = ");
      if (debug == 1) Serial.print(key);
      if (debug == 1) Serial.print("   val = ");
      if (debug == 1) Serial.println(number);
    }
  }
  while (counter <= 3); //  collect exactly 4 digits
  LocoAddress[ActiveAddress] = total;
  saveAddresses();
  tft.fillScreen(TFT_BLACK);
  doMainLCD();
}  // END  getLocoAddress

void doDCC() {
  if (debug == 1) Serial.println(LocoDirection[ActiveAddress] );
//  Serial.print("<1>");
  Serial.print("<t1 ");
  Serial.print(LocoAddress[ActiveAddress] );//locoID);
  Serial.print(" ");
  Serial.print(LocoSpeed[ActiveAddress] );
  Serial.print(" ");
  Serial.print(LocoDirection[ActiveAddress] );
  Serial.println(">");
}  // END doDCC

void doDCCfunction04() {
  Serial.write("<f ");
  Serial.print(LocoAddress[ActiveAddress] );
  Serial.print(" ");
  int fx = LocoFN0to4[ActiveAddress] + 128 ; 
  Serial.print(fx);
  Serial.print(" >");
}

void doDCCfunction58() {
  Serial.write("<f ");
  Serial.print(LocoAddress[ActiveAddress] );
  Serial.print(" ");
  int fx = LocoFN5to8[ActiveAddress] + 176;
  Serial.print(fx);
  Serial.print(" >");
}

void doDCCfunction912() {         //160 = F9-F12
  Serial.write("<f ");
  Serial.print(LocoAddress[ActiveAddress] );
  Serial.print(" ");
  int fx = LocoFN9to12[ActiveAddress] + 160;
  Serial.print(fx);
  Serial.print(" >");
}

void doDCCfunction1320() {
  Serial.write("<f ");
  Serial.print(LocoAddress[ActiveAddress] );
  Serial.print(" 222 ");
  int fx = LocoFN13to20[ActiveAddress];
  Serial.print(fx);
  Serial.print(" >");
}

void doDCCfunction2128() {
  Serial.write("<f ");
  Serial.print(LocoAddress[ActiveAddress] );
  Serial.print(" 223 ");
  int fx = LocoFN21to28[ActiveAddress];
  Serial.print(fx);
  Serial.print(" >");
}

void doDCCpoint() { 
//  Serial.print ("<1>");  
  Serial.print ("<a ");
  Serial.print ((( PointAddressDble / 2 - 1 ) / 4) + 1 ); // Point Address, DCC++ works with address + sub address
  Serial.print (" ");
  Serial.print (( PointAddressDble / 2  - 1 ) % 4 );  // Point Sub Address 
  if (PointAddressDble % 2 ) {
    Serial.println (" 0 >");
  }
  else {
    Serial.println (" 1 >"); // Throw - dévié
  }
}

void doDCCaccessory() { 
//  Serial.print ("<1>");  
  Serial.print ("<a ");
  Serial.print ( ( (AccessoryAddressDble  / 2 - 1 ) / 4) + 1 ); // Accessory Address, DCC++ works with address + sub address
  Serial.print (" ");
  Serial.print (( AccessoryAddressDble / 2  - 1 ) % 4 );  // Accessory Sub Address 
  if (AccessoryAddressDble % 2 ) {
    Serial.println (" 0 >");  // desactive
  }
  else {
    Serial.println (" 1 >");  // active
  }
}

void doDCCturntable() { 
//  Serial.print ("<1>");  
  Serial.print ("<a 17 1");     //  address 65 for motor relays, activation for :
  if (TurntableAddressDble % 2 ) {
    Serial.println (" 0 >");    // clockwise
  }
  else {
    Serial.println (" 1 >");    // counter clockwise
  }
  Serial.print ("<a 17 2 1>");  // address 66 for solenoid, 200 ms duration
  delay (200);
  Serial.print ("<a 17 2 0>");  // solenoid for 200 ms
}

void all2ZeroSpeeed() {  // set flag to 1 to stop, set to 0 to restore
  for (int tempx = 0; tempx <= maxLocos; tempx++) {
    Serial.print("<t1 ");
    Serial.print(LocoAddress[tempx] );//locoID);
    Serial.print(" ");
    if (ZeroSpeedFlag == 1) {
      Serial.print(0);//LocoSpeed[0] );
    }
    else Serial.print(LocoSpeed[0] );
    Serial.print(" ");
    Serial.print(LocoDirection[1] );
    Serial.write(">");
  }
}

void getAddresses() {
  EEPROM.get(0, ms);
  EEPROM.commit();
  LocoAddress[0]= ms.address1;
  LocoAddress[1]= ms.address2;
  LocoAddress[2]= ms.address3;
  LocoAddress[3]= ms.address4;
  maxLocos = ms.struct_version;
  if (maxLocos >= 4) maxLocos = 4;
}

void saveAddresses() {
  switch (ActiveAddress){
    case 0: 
      ms.address1 =  LocoAddress[ActiveAddress] ;
      break;
    case 1:    
      ms.address2 =  LocoAddress[ActiveAddress] ;
      break;
    case 2:    
      ms.address3 =  LocoAddress[ActiveAddress] ;
      break;
    case 3:      
      ms.address4 =  LocoAddress[ActiveAddress] ;
      break;    
  }
  ms.struct_version = maxLocos ;
  EEPROM.put(0, ms);
  EEPROM.commit();
//  getAddresses();
}

void doMainLCD() {
  tft.setCursor(0,0); // 24,0
  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  tft.print("V= ");
  tft.setTextColor(TFT_YELLOW, TFT_BLACK);
  tft.print(LocoSpeed[ActiveAddress], DEC); // 
  tft.print("     ");     // 
  tft.setCursor(80,0);
//  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  if (LocoDirection[ActiveAddress] == 1 ) {
    tft.print(">>> ");
  }
  else {
    tft.print("<<< ");
  }
  tft.setCursor(0, 20); // 2nd line
  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  tft.print("L= ");
  tft.setTextColor(TFT_GREEN, TFT_BLACK);
  if (LocoAddress[ActiveAddress] <= 9) {
    tft.print("0");  // add leading zero for single digit addresses
  }
  tft.print(LocoAddress[ActiveAddress] , DEC);
  tft.print("   "); 
  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  tft.setCursor(96, 20); // pre last char line 2
  tft.print("#");
  tft.setTextColor(TFT_GREEN, TFT_BLACK);
  tft.setCursor (108, 20);       // last char line 2
  tft.print(ActiveAddress + 1);
  tft.setTextColor(TFT_WHITE, TFT_BLACK);  
  int fonct = LocoFN0to4[ActiveAddress] ;// 
  tft.setCursor(0, 40); 
  for (int i = 4 ; i >= 0 ; i--) { ///
    tft.setCursor(12 + (12 * i) , 40); ///
    tft.print( bitRead (fonct, i)); 
  }
  tft.setCursor(60, 40); 
  tft.print(" "); 
  tft.setCursor(0, 40); 
  tft.print( bitRead (fonct, 4)); // F0 ///
  fonct = LocoFN5to8[ActiveAddress] ;// 
  for (int i =0 ; i< 4; i++) {
    tft.setCursor(72 + (12 * i), 40); // 3e G0
    tft.print( bitRead (fonct, i)); 
  }
//  tft.print(" "); 
  fonct = LocoFN9to12[ActiveAddress] ;// 
  for (int i =0 ; i< 4; i++) {
    tft.setCursor(12 * i, 60); // 4e G1
    tft.print( bitRead (fonct, i)); 
  }
  tft.print(" "); 
  fonct = LocoFN13to20[ActiveAddress] ;// 
  for (int i =0 ; i< 8; i++) {
    tft.setCursor(12 * i, 80); // 5e
    tft.print( bitRead (fonct, i)); 
  }
  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  tft.print(" 2"); 
  tft.setTextColor(TFT_WHITE, TFT_BLACK);
  fonct = LocoFN21to28[ActiveAddress] ;//
  for (int i =0 ; i< 8; i++) {
    tft.setCursor(12 * i, 100); //6e
    tft.print( bitRead (fonct, i)); 
  }
  tft.setTextColor(TFT_CYAN, TFT_BLACK);
  tft.print(" 3");       
  tft.setTextColor(TFT_WHITE, TFT_BLACK);

}  // END doMainLCD

void doPointLCD() {
  tft.setCursor(0,0); // 
  tft.setTextColor(TFT_CYAN, TFT_BLACK);  
  tft.println("Aiguillage");
  tft.setTextColor(TFT_WHITE, TFT_BLACK);  
  tft.setCursor(0,20); // // line 2
  tft.print(PointAddressDble / 2 , DEC); 
  tft.print("         ");
  tft.setCursor(0, 40); // line 3
  tft.print(((PointAddressDble / 2 - 1) / 4) + 1, DEC ); // main address
  tft.print(" / ");  
  tft.print(((PointAddressDble / 2 - 1) % 4 ), DEC );  // sub address
  tft.print("     ");    
  // tft.setCursor(96, 40); // before last character line 3
  // tft.print("#");
  tft.setCursor(0, 60); // start of 4th line 
  if (PointAddressDble % 2 ) {
    tft.setTextColor(TFT_GREEN, TFT_BLACK);
    tft.print("Devie      ");
    tft.setTextColor(TFT_WHITE, TFT_BLACK);
  }
  else {
    tft.setTextColor(TFT_YELLOW, TFT_BLACK);
    tft.print("Normal     ");
    tft.setTextColor(TFT_WHITE, TFT_BLACK);
  }
  tft.setCursor(0, 80); // start of 5th line
  tft.print("Appuyer       ");      
  tft.setCursor(0, 100); // start of 6th line
  tft.print("              ");      
}  // END doPointLCD

void doAccessoryLCD() {
  tft.setCursor(0,0); // 
  tft.setTextColor(TFT_CYAN, TFT_BLACK);  
  tft.println("Accessoire");
  tft.setTextColor(TFT_WHITE, TFT_BLACK);  
  tft.setCursor(0,20); // 
  tft.print(AccessoryAddressDble / 2, DEC); 
//  tft.print(((AccessoryAddressDble / 4 ) - 3 ) , DEC );
  tft.print("         ");
  tft.setCursor(0, 20); // line 2
/*  tft.print(((AccessoryAddressDble  - 1) / 2) + 1, DEC ); // main address
  tft.print(" / ");  
  tft.print(((AccessoryAddressDble  - 1) % 2 ), DEC );  // sub address
  tft.print("  ");    
  tft.setCursor(96, 20); // before last character line 2
*/
//  tft.print("         #");
  tft.setCursor(0, 40); // start of 3rd line 
  if (AccessoryAddressDble % 2 ) {
    tft.setTextColor(TFT_YELLOW, TFT_BLACK);
    tft.print("Desactiver ");
    tft.setTextColor(TFT_WHITE, TFT_BLACK);
  }
  else {
    tft.setTextColor(TFT_GREEN, TFT_BLACK);
    tft.print("Activer     ");
    tft.setTextColor(TFT_WHITE, TFT_BLACK);
  }
  tft.setCursor(0, 60); // start of 4th line
  tft.print("Appuyer       ");      
  tft.setCursor(0, 80); // start of 5th line
  tft.print("              ");      
  tft.setCursor(0, 100); // start of 6th line
  tft.print("              ");      
}  // END doAccessoryLCD

void doTurntableLCD() {
  tft.setCursor(0,0); // 
  tft.setTextColor(TFT_CYAN, TFT_BLACK);  
  tft.print("Plaque T. ");
  tft.setTextColor(TFT_WHITE, TFT_BLACK);  
  tft.setCursor(0,20); //
  tft.print("sens :     "); 
  //  tft.print(TurntableAddressDble / 2, DEC); 
  tft.setCursor(0, 40); // start of 3rd line 
  if (TurntableAddressDble % 2 ) {
    tft.setTextColor(TFT_YELLOW, TFT_BLACK);
    tft.print("Horaire    ");
    tft.setTextColor(TFT_WHITE, TFT_BLACK);
  }
  else {
    tft.setTextColor(TFT_GREEN, TFT_BLACK);
    tft.print("Anti-horaire");
    tft.setTextColor(TFT_WHITE, TFT_BLACK);
  }
  tft.setCursor(0, 60); // start of 4th line
  tft.print("Appuyer       ");      
  tft.setCursor(0, 80); // start of 5th line
  tft.print("              ");      
  tft.setCursor(0, 100); // start of 6th line
  tft.print("              ");  
}
